__author__ = 'nick'
import numpy as np
import scipy.integrate
from scipy.stats import beta

class DiagStore(object):
    def __init__(self):
        pass


def inv_logit(x): return 1 / (1 + np.exp(-x))


class BayesianModel(object):

    def __init__(self, gamma_0, prior, n_stimuli):
        self.gamma_0 = gamma_0
        self.gamma = 1 / (1 + np.exp(-gamma_0))
        self.prior = prior
        self.n_stimuli = n_stimuli

        self.alpha = np.ones((2, n_stimuli)) * prior
        self.beta = np.ones((2, n_stimuli)) * prior

        self.stimulus_counter = np.zeros(n_stimuli)

    def select_action(self, stimulus):
        a = np.random.beta(self.alpha[0, stimulus], self.beta[0, stimulus]) < \
            np.random.beta(self.alpha[1, stimulus], self.beta[1, stimulus])

        return a

    def sample_as_prob(self, stimulus, n_samples):
        """
        :return Probability (float) action 0 is selected given stimulus
        """
        v = np.random.beta(self.alpha[0, stimulus], self.beta[0, stimulus], n_samples)
        k = np.random.beta(self.alpha[1, stimulus], self.beta[1, stimulus], n_samples)
        return np.mean(v > k)

    def update(self, stimulus, action,  reward):

        # decay counts before updating
        self.alpha[:, stimulus] *= self.gamma
        self.beta[:, stimulus] *= self.gamma

        # update counts
        self.alpha[action, stimulus] += int(reward == 1)
        self.beta[action, stimulus] += int(reward != 1)

        # print np.sum(self.alpha), np.sum(self.beta)

        # print self.alpha
        # print self.beta

    def get_diagnostics(self):
        diags = DiagStore()
        diags.alpha = self.alpha
        diags.beta = self.beta
        diags.pr0 = self.sample_as_prob(0, 250)
        diags.pr1 = self.sample_as_prob(1, 250)
        diags.h0 = self.get_entropy(0)
        diags.h1 = self.get_entropy(1)
        return diags

    def get_entropy(self, stimulus):
        """
        :param stimulus:
        :return:
        """
        a0 = self.alpha[0, stimulus]
        a1 = self.alpha[1, stimulus]
        b0 = self.beta[0, stimulus]
        b1 = self.beta[1, stimulus]

        fn = lambda x: beta.pdf(x, a0, b0) * beta.cdf(x, a1, b1)
        pr = scipy.integrate.quad(fn, 0, 1)[0]

        H = (-pr * np.log(pr)) - ((1-pr) * np.log(1-pr))

        return H

    def likelihood(self):
        pass


class BayesianModel_wSlope(BayesianModel):

    def __init__(self, gamma_0, gamma_1, eta, prior, n_stimuli):
        self.gamma_0 = gamma_0
        self.gamma_1 = gamma_1
        self.eta = eta
        self.prior = prior
        self.n_stimuli = n_stimuli

        self.alpha = np.ones((2, n_stimuli)) * prior
        self.beta = np.ones((2, n_stimuli)) * prior

        self.stimulus_counter = np.zeros(n_stimuli)

        self.h_state = np.ones(n_stimuli)
        self.smoothed_h_slope = np.zeros(n_stimuli)

    def select_action(self, stimulus):
        a = np.random.beta(self.alpha[0, stimulus], self.beta[0, stimulus]) < \
            np.random.beta(self.alpha[1, stimulus], self.beta[1, stimulus])

        return a

    def sample_as_prob(self, stimulus, n_samples):
        """
        :return Probability (float) action 0 is selected given stimulus
        """
        v = np.random.beta(self.alpha[0, stimulus], self.beta[0, stimulus], n_samples)
        k = np.random.beta(self.alpha[1, stimulus], self.beta[1, stimulus], n_samples)
        return np.mean(v > k)

    def update(self, stimulus, action,  reward):

        # Get the smoothed entropy slope
        h = self.get_entropy(stimulus)
        slope = h - self.h_state[stimulus]
        self.h_state[stimulus] = h # store current entropy for next encounter of stim
        self.smoothed_h_slope[stimulus] += self.eta * (slope - self.smoothed_h_slope[stimulus])

        # print self.smoothed_h_slope[stimulus]
        gamma = self.gamma_0 + self.gamma_1 * np.float(self.smoothed_h_slope[stimulus])
        gamma = 1 / (1 + np.exp(-gamma))

        # decay counts before updating
        # self.alpha[:, stimulus] = gamma * self.alpha[:, stimulus] + 1/gamma - 1
        # self.beta[:, stimulus] = gamma * self.beta[:, stimulus] + 1/gamma - 1

        self.alpha[:, stimulus] *= gamma
        self.beta[:, stimulus] *= gamma

        # self.alpha[:, stimulus] += 1e-4
        # self.beta[:, stimulus] += 1e-4

        # print self.alpha
        # print self.beta

        # update counts
        self.alpha[action, stimulus] += int(reward == 1)
        self.beta[action, stimulus] += int(reward != 1)

    def get_diagnostics(self):
        diags = DiagStore()
        diags.alpha = self.alpha
        diags.beta = self.beta
        diags.pr0 = self.sample_as_prob(0, 250)
        diags.pr1 = self.sample_as_prob(1, 250)
        diags.h0 = self.get_entropy(0)
        diags.h1 = self.get_entropy(1)
        # diags.h0_slope = []
        # diags.h1_slope = []
        return diags

    def likelihood(self):
        pass


class BayesianModel_w_noise(BayesianModel_wSlope):

    def __init__(self, gamma_0, gamma_1, eta, prior, noise_mu, noise_sigma, n_stimuli):

        BayesianModel_wSlope.__init__(self, gamma_0, gamma_1, eta, prior, n_stimuli)

        self.noise_mu = noise_mu
        self.noise_sigma = noise_sigma

    def update(self, stimulus, action,  reward):

        # Get the smoothed entropy slope
        h = self.get_entropy(stimulus)
        slope = h - self.h_state[stimulus]
        self.h_state[stimulus] = h # store current entropy for next encounter of stim
        self.smoothed_h_slope[stimulus] += self.eta * (slope - self.smoothed_h_slope[stimulus])

        # print self.smoothed_h_slope[stimulus]
        gamma = self.gamma_0 + self.gamma_1 * np.float(self.smoothed_h_slope[stimulus])
        gamma = 1 / (1 + np.exp(-gamma))

        # decay counts before updating
        # self.alpha[:, stimulus] = gamma * self.alpha[:, stimulus] + 1/gamma - 1
        # self.beta[:, stimulus] = gamma * self.beta[:, stimulus] + 1/gamma - 1

        self.alpha[:, stimulus] *= gamma
        self.beta[:, stimulus] *= gamma

        #draw the trial-by-trial noise on the 
        p = inv_logit(np.random.normal(self.noise_mu, self.noise_sigma))

        # update counts
        self.alpha[action, stimulus] += int(reward == 1) * (1-p)
        self.beta[action, stimulus] += int(reward != 1) * (1-p)

        # update the counts for alternate aciton with random noise
        self.alpha[~action, stimulus] += int(reward == 1) * p
        self.beta[~action, stimulus] += int(reward != 1) * p

def generative_run_agent(myagent, stims, rs, trials_per_epoch=20):
    """
    Generative Model:

    Usage: [Choice,Reward,probe] = generative_run_agent(agent ,stims, rs)


    :param myagent:
    :param stims:
    :param rs:
    :return:
    """

    n = len(stims)
    n_s = len(set(stims))

    # [h, pr] = #todo: store entropy

    # DATA:
    # 1: choice (0,1)
    # 2: reward (0,1)

    # ~~~ Begin ~~~~
    # alpha = np.ones((n_s, 2)) * initialization_parameter
    # beta = np.ones((n_s, 2)) * initialization_parameter

    choices = np.zeros(n)
    rewards = np.zeros(n)
    choices[:] = np.nan
    rewards[:] = np.nan
    probe0 = list()
    probe1 = list()

    # counters
    pb = 0
    t = np.zeros(2)

    for ii in range(n):

        if (ii % (n_s*20)) == 0:
            probe0.append(myagent.sample_as_prob(0, 250))
            probe1.append(myagent.sample_as_prob(1, 250))

        s = stims[ii]
        t[s] += 1

        # Have the agent select the action
        a = myagent.select_action(s)
        # print a
        choices[ii] = a

        # Is it rewarded?
        r = a == rs[ii]
        rewards[ii] = r

        # Todo: entropy determine model entropy!

        # Update agent
        myagent.update(s, a, r)

    probe = (np.asarray(probe0), np.asarray(probe1))

    return choices, rewards, probe




def debug_generative_run_agent(agent, stims, rs, trials_per_epoch=20):
    """
    Generative Model:

    Usage: [Choice,Reward,probe] = generative_run_agent(agent ,stims, rs)


    :param agent:
    :param stims:
    :param rs:
    :return:
    """

    n = len(stims)
    n_s = len(set(stims))

    # [h, pr] = #todo: store entropy

    # DATA:
    # 1: choice (0,1)
    # 2: reward (0,1)

    # ~~~ Begin ~~~~
    # alpha = np.ones((n_s, 2)) * initialization_parameter
    # beta = np.ones((n_s, 2)) * initialization_parameter

    choices = np.zeros(n)
    rewards = np.zeros(n)
    choices[:] = np.nan
    rewards[:] = np.nan
    probe = list()

    # counters
    pb = 0
    t = np.zeros(2)
    diagnostics = list()

    for ii in range(n):

        diagnostics.append(agent.get_diagnostics())

        if (ii % (n_s*20)) == 0:
            probe.append((agent.sample_as_prob(0, 250), agent.sample_as_prob(1, 250)),)

        s = stims[ii]
        t[s] += 1

        # Have the agent select the action
        a = agent.select_action(s)
        # print a
        choices[ii] = a

        # Is it rewarded?
        r = a == rs[ii]
        rewards[ii] = r

        # Todo: entropy determine model entropy!

        # Update agent
        agent.update(s, a, r)

    return diagnostics, probe
